#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterator, Optional, Sequence, Type, TypeVar, Union, Tuple, Any, List, Dict, Set, cast, overload
import builtins


#--------------------------------------------------------------------------------
# 전역 상수 목록.
#--------------------------------------------------------------------------------
BASE : str = "base"


#--------------------------------------------------------------------------------
# 빌트인 확장 클래스.
# - 파이썬의 기본 시스템 내장 함수들은 빌트인 모듈로 접근하여 사용 가능.
# - bultins.print()와 print()은 동일 함수.
# - hasattr, getattr, setattr, delattr, vars, dir.
#--------------------------------------------------------------------------------
class Builtins:
	#--------------------------------------------------------------------------------
	# 출력.
	#--------------------------------------------------------------------------------
	@staticmethod
	def Print(*values : str) -> None:
		builtins.print(*values)


	#--------------------------------------------------------------------------------
	# 대상이 타입 인스턴스인지 여부 반환.
	# - instance = Hello()
	# - instance : 인스턴스
	# - instance.__class__ : 인스턴스의 클래스 타입 (class Hello)
	# - instance.__class__ == Hello : 동일
	#
	# - instance.__class__.__class__ : 클래스의 클래스 타입 (class type)
	# - instance.__class__.__class__ == type : 동일
	# - 타입은 "클래스의 메타클래스"로서 클래스는 타입의 인스턴스라고 할 수 있다.
	# - instance < class < type
	# - instance는 type과 직접적인 관련이 없기 때문에 같지 않다.
	#--------------------------------------------------------------------------------
	@staticmethod
	def IsType(obj : object) -> bool:
		if not builtins.isinstance(obj, builtins.type):
			return False
		return True


	#--------------------------------------------------------------------------------
	# 대상이 클래스 타입에 해당하는 인스턴스인지 여부 반환.
	#--------------------------------------------------------------------------------
	@staticmethod
	def IsInstanceType(obj : object, classType : Type) -> bool:
		if not builtins.isinstance(obj, classType):
			return False
		return True


	#--------------------------------------------------------------------------------
	# 대상이 클래스 타입 목록 중 하나에 해당하는 인스턴스인지 여부 반환.
	#--------------------------------------------------------------------------------
	@staticmethod
	def IsInstanceInTypes(obj : object, classTypes : List[Type]) -> bool:
		if not builtins.isinstance(obj, tuple(classTypes)):
			return False
		return True


	#--------------------------------------------------------------------------------
	# 실제 어트리뷰트 이름 반환.
	# - 존재한다면 실제 어트리뷰트 이름을 반환. (입력값과 동일할 수도 다를 수도 있음)
	# - 존재하지 않으면 빈문자열 반환.
	#--------------------------------------------------------------------------------
	@staticmethod
	def GetRealAttributeName(obj : object, attributeName : str) -> str:
		if not obj:
			return str()
		if not attributeName:
			return str()
		attributeNameLower : str = attributeName.lower()
		for realAttributeName in builtins.dir(obj):
			realAttributeNameLower : str = realAttributeName.lower()
			if attributeNameLower == realAttributeNameLower:
				return realAttributeName				
		return str()


	#--------------------------------------------------------------------------------
	# 어트리뷰트 설정. (혹은 생성)
	#--------------------------------------------------------------------------------
	@staticmethod
	def SetAttribute(obj : object, attributeName : str, value : Any, allowCreation : bool = True) -> bool:
		realAttributeName = Builtins.GetRealAttributeName(obj, attributeName)
		if realAttributeName:
			builtins.setattr(obj, attributeName, value)
			return True
		elif allowCreation:
			builtins.setattr(obj, attributeName, value)
			return True
		else:
			return False


	#--------------------------------------------------------------------------------
	# 어트리뷰트 제거.
	#--------------------------------------------------------------------------------
	@staticmethod
	def DeleteAttribute(obj : object, attributeName : str) -> bool:
		realAttributeName = Builtins.GetRealAttributeName(obj, attributeName)
		if not realAttributeName:
			return False
		
		builtins.delattr(realAttributeName)
		return True


	#--------------------------------------------------------------------------------
	# 어트리뷰트 반환.
	#--------------------------------------------------------------------------------
	@staticmethod
	def GetAttribute(obj : object, attributeName : str, defaultValue : Any = None) -> Any:
		if not obj:
			return defaultValue
		if not attributeName:
			return defaultValue
		realAttributeName : str = Builtins.GetRealAttributeName(obj, attributeName)
		if not realAttributeName:
			return defaultValue
		return builtins.getattr(obj, realAttributeName)


	#--------------------------------------------------------------------------------
	# 어트리뷰트의 존재 유무 판단.
	#--------------------------------------------------------------------------------
	@staticmethod
	def HasAttribute(obj : object, attributeName : str) -> bool:
		if not obj:
			return False
		if not attributeName:
			return False
		realAttributeName : str = Builtins.GetRealAttributeName(obj, attributeName)
		if not realAttributeName:
			return False
		return True


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterator, Optional, Sequence, Type, TypeVar, Union, Tuple, Any, List, Dict, Set, cast, overload
import inspect
from functools import wraps
from .core import Builtins



#--------------------------------------------------------------------------------
# 비공개 메서드를 위한 오버라이드 메서드 데코레이터 (인스턴스메서드 전용).
# - 동일 이름의 자식이 있으면 자식을 호출.
# - 동일 이름의 자식이 없으면 자신을 호출.
# - 최상위 메서드에서만 지정.
#--------------------------------------------------------------------------------
def overridemethod(targetMethod : Callable[..., Any]):
	@wraps(targetMethod)
	def Decorate(self, *args, **kwargs) -> Any:
		# 인자 없음 - 일반 함수 혹은 스태틱메서드 일 경우.
		if not args: return targetMethod(*args, **kwargs)

		argument = args[0]

		# 인자 있음 - 하지만 현재 호출한 메서드의 인스턴스나 현재 호출한 메서드의 클래스가 아닐 경우.
		# if not isinstance(argument, type) and not isinstance(argument, self.__class__): return targetMethod(*args, **kwargs)
		if not Builtins.IsInstanceInTypes(argument, [type, self.__class__]): return targetMethod(*args, **kwargs)

		# 첫번째 인자가 클래스 타입 일 경우 or 인스턴스 일 경우.
		isClassMethod = Builtins.IsType(argument, type)
		if isClassMethod:
			classType = args[0]
		else:
			classType = args[0].__class__
	
		# 메서드 이름 룰이 비공개 메서드 일 경우 or 공개 메서드 일 경우.
		methodName = targetMethod.__name__
		isPrivateMethod = methodName.startswith("__") and not methodName.endswith("__")
		if isPrivateMethod:
			methodName = methodName.split("__", 1)[-1]
			methodName = f"_{classType.__name__}__{methodName}"

		# 자식 에게 동일 이름의 함수가 있다면 호출.
		if Builtins.HasAttribute(classType, methodName):
			childMethod = Builtins.GetAttribute(classType, methodName)
			if childMethod is not targetMethod:
				return childMethod(*args, **kwargs)
			
		# 없으면 자신의 함수 호출.
		return targetMethod(*args, **kwargs)
	return Decorate


#--------------------------------------------------------------------------------
# 자식의 메소드가 오버라이드 한 것일 때 부모의 동일한 메소드를 대신 호출 해주는 함수.
#--------------------------------------------------------------------------------
def basemethod(self, *args, **kwargs) -> Any:
	currentFrame = inspect.currentframe()
	previouslyFrame = inspect.getouterframes(currentFrame)[1]
	methodName = previouslyFrame.function
	parentClass = self.__class__.__bases__[0]

	# 비공개 메서드 이름 장식 처리
	isPrivateMethod = methodName.startswith("__") and not methodName.endswith("__")
	if isPrivateMethod:
		methodName = methodName.split("__", 1)[-1]
		methodName = f"_{parentClass.__name__}__{methodName}"

	# 부모 클래스의 메서드를 가져와 호출
	if hasattr(parentClass, methodName):
		parentMethod = getattr(parentClass, methodName)
		return parentMethod(self, *args, **kwargs)
	else:
		raise AttributeError(f"{parentClass.__name__} object has no attribute {methodName}")



#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterator, Optional, Sequence, Type, TypeVar, Union, Tuple, Any, List, Dict, Set, cast, overload
import builtins


#--------------------------------------------------------------------------------
# 패키지 안의 클래스 별칭 목록.
#--------------------------------------------------------------------------------
from .anonymousclass import AnonymousObject, UnnamedClass
from .baseclass import BaseClass
from .baseclass import BaseClass as Object
from .baseconstant import BaseConstant
from .baseconstant import BaseConstant as Constant
from .basemetaclass import BaseMetaClass
from .basemetaclass import BaseMetaClass as MetaClass
from .basemetaclass import BaseMetaClass as Meta
from .basenode import BaseNode
from .basenode import BaseNode as Node
from .builtins import Builtins


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterator, Optional, Sequence, Type, TypeVar, Union, Tuple, Any, List, Dict, Set, cast, overload
import platform
from enum import Enum, auto
import subprocess
from .core import Builtins


#--------------------------------------------------------------------------------
# 플랫폼 타입.
#--------------------------------------------------------------------------------
class PlatformType(Enum):
	UNKNOWN = auto()
	
	# windows.
	WINDOWS = auto()

	# linux.
	LINUX = auto()
	ANDROID = auto()
	STEAMOS = auto()

	# apple.
	MACOS = auto()
	IOS = auto()
	IPADOS = auto()
	WATCHOS = auto()
	TVOS = auto()
	VISIONOS = auto()

	# web.
	WEBGL = auto()
	WEBGPU = auto()


#--------------------------------------------------------------------------------
# 현재 시스템의 플랫폼 타입 반환.
#--------------------------------------------------------------------------------
def GetPlatformType() -> PlatformType:
	systemName : str = platform.system()
	systemName = systemName.upper()
	if systemName == "WINDOWS":
		return PlatformType.WINDOWS
	elif systemName == "DARWIN":
		return PlatformType.MACOS
	elif systemName == "LINUX":
		return PlatformType.LINUX
	else:
		return PlatformType.UNKNOWN
	

#--------------------------------------------------------------------------------
# 현재 시스템에 설치된 파이썬 목록을 별도 콘솔에서 조회하여 반환.
#--------------------------------------------------------------------------------
def GetPythonInterpreters() -> list[str]:
	try:
		platformType : PlatformType = GetPlatformType()
		if platformType == PlatformType.WINDOWS:
			output : str = subprocess.check_output("where python", shell = True, text = True)
		else:
			output : str = subprocess.check_output("which -a python python3", shell = True, text = True)
		interpreters : list[str] = output.strip().split("\n")
		return interpreters
	except Exception as exception:
		return list()


#--------------------------------------------------------------------------------
# 참조 모듈 목록.
#--------------------------------------------------------------------------------
from __future__ import annotations
from typing import Awaitable, Callable, Final, Generic, Iterator, Optional, Sequence, Type, TypeVar, Union, Tuple, Any, List, Dict, Set, cast, overload
import pkgutil


#--------------------------------------------------------------------------------
# 패키지 안의 클래스 별칭 목록.
#--------------------------------------------------------------------------------
from .core import AnonymousObject, UnnamedClass
from .core import BaseClass, Object
from .core import BaseConstant, Constant
from .core import BaseMetaClass, MetaClass, Meta
from .core import BaseNode, Node
from .core import Builtins
# from .deprecated import BaseNode
from .exception import SingletonException
from .filesystem import Entry, Directory, Drive, File, Storage
# from .future import EventNode, Node
from .manager import BaseRepository, Repository, SharedClass, Singleton
from .decorator import overridemethod, basemethod
from .enumflag import EnumFlag, auto
from .environment import PlatformType, GetPlatformType
from .path import Path